#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>


#if HAVE_STRINGS_H
#include <strings.h>
#endif

#include "irrdi.h"
#include "trace.h"
#include "utils.h"

char conn_err[512];

extern int debug;

#ifndef HAVE_HSTRERROR

const char*
hstrerror(int err)
{ 
	static char  buffer[512];
	snprintf(buffer,sizeof(buffer),"resolver error %i",err);
	return buffer;
};
#endif

#ifndef HAVE_GETNAMEINFO
int
fill_sca(char* host,int port,struct sockaddr_in* sca)
{ 
	struct hostent* hte;

	bzero(conn_err,sizeof(conn_err));
	bzero(sca,sizeof(struct sockaddr_in));

	hte=gethostbyname(host);
	if(!hte) { 
		snprintf(conn_err,sizeof(conn_err),"gethostbyname: %s",hstrerror(h_errno));
		return h_errno;
	};

	bcopy(hte->h_addr_list[0],&sca->sin_addr.s_addr,hte->h_length);
	sca->sin_family=PF_INET;
	sca->sin_port=htons(port);

	return 0;
};
#endif

struct irrdc* 
irrd_init(char* host, int port)
{ 
	struct irrdc* ir=malloc(sizeof(struct irrdc));
	if(!ir) return NULL;
	bzero(ir,sizeof(struct irrdc));
	if(host) ir->host=strdup(host);
	if(port) ir->port=port;
	ir->s=-1;

	if(host&&port) { 
#if !HAVE_GETADDRINFO
		if(fill_sca(host,port,&ir->sca)) { 
			SX_TRACE2("%s",conn_err);
			if(ir->host) free(ir->host);
			free(ir);
			return NULL;
		};
#else
		int gerror;
		struct addrinfo hints;
		char servname[15];
		memset(&hints,0,sizeof(struct addrinfo));
		hints.ai_socktype=SOCK_STREAM;
		hints.ai_protocol=IPPROTO_TCP;
		snprintf(servname,sizeof(servname),"%i",port);
		if((gerror=getaddrinfo(host,servname,&hints,&ir->res))) { 
			char buffer[2048];
			snprintf(buffer,sizeof(buffer),"getaddrinfo(%s) failed: %s",
				host,gai_strerror(gerror));
			SX_TRACE2("%s",buffer);
			if(ir->host) free(ir->host);
			free(ir);
			return NULL;
		};
#endif
			
	};
	return ir;
};

int
irrd_connect(struct irrdc* ir)
{ 
	if(ir->s) close(ir->s);
	ir->s=-1;
#if !HAVE_GETADDRINFO
	ir->s=socket(AF_INET,SOCK_STREAM,0);
	if(ir->s==-1) { 
		return 0;
	};
	if(debug>2) { 
		SX_TRACE2("Connecting to %s...",inet_ntoa(ir->sca.sin_addr));
	};
	if(connect(ir->s,(struct sockaddr*)&ir->sca,sizeof(struct sockaddr_in))==-1) { 
		return 0;
	};
#else 
	{ 
		struct addrinfo *ress=ir->res;
		for(;ress;ress=ress->ai_next) { 
			if(debug>=2) { 
				char hostname[64], service[32];
				getnameinfo(ress->ai_addr,ress->ai_addrlen,hostname,
					sizeof(hostname), service, sizeof(service), 
					NI_NUMERICHOST|NI_NUMERICSERV);
				SX_TRACE2("Connecting to %s:%s...", hostname, service);
			};
			ir->s=socket(ress->ai_family,ress->ai_socktype,ress->ai_protocol);
			if(ir->s<0) { 
				SX_TRACE2("Unable to create socket: %s", strerror(errno));
				continue;
			};
			if(connect(ir->s,ress->ai_addr,ress->ai_addrlen)<0) { 
				SX_TRACE2("Unable to connect: %s", strerror(errno));
				close(ir->s);
				ir->s=-1;
				continue;
			};

			break;
		};
		if(ir->s==-1) { 
			return 0;
		};
	};
#endif
	if(debug>2) { 
		SX_TRACE2("sending !!");
	};
	write(ir->s, "!!\n",3);

	return 1;
};

int
irrd_question(struct irrdc* ir, char* quest, ...)
{
	va_list ap;
	char qbuff[512];
	char rbuff[2048];
	char *p;
	size_t len, n;
	int need_copy=0;
	int retries = 0;

	memset(qbuff,0,sizeof(qbuff));
	va_start(ap,quest);
	
	vsnprintf(qbuff,sizeof(qbuff),quest,ap);

	va_end(ap);

	while (1) {
	    if(ir->answer) free(ir->answer);
	    ir->answer=NULL;

	    if(debug>2) 
		    SX_TRACE2("sending %s", qbuff);

	    if(!retries)
		strcat(qbuff, "\n");
	    if (write(ir->s, qbuff, strlen(qbuff)) != strlen(qbuff))
		    goto errlab;

	    if((n = read(ir->s, rbuff, sizeof(rbuff)))<=0)
		    goto errlab;
	    
	    p = strchr(rbuff, '\n');

	    if (!p) 
		goto errlab;

	    *p++ = '\0';
	    if(n > strlen(rbuff)+1)
		need_copy=1;

	    if(debug>2) 
		    SX_TRACE2("Answer is %s", rbuff);

	    switch(rbuff[0]) { 
		    case 'A': 
			    len=atoi(rbuff+1);
			    ir->answer=malloc(len+1);
			    if(!ir->answer) { 
				    return 0;
			    };
			    bzero(ir->answer,len+1);

			    if(debug>2) 
				    SX_TRACE2("Waiting for %i data bytes",len);

			    if(need_copy) {
				n -= strlen(rbuff)+1;
				if(n > len) {
				    need_copy = n - len;
				    n = len;
				} else
				    need_copy=0;
				bcopy(p, ir->answer, n);
				p+=n;
				if(debug>2)
				    SX_TRACE2("Copy %d bytes", n);
			    } else
				n = 0;

			    while(n < len) {
				size_t nn;
				nn = read(ir->s, ir->answer + n, len-n);
				if (nn<=0)
				    goto errlab;
				n += nn;
			    }
			    
			    if(debug>2) 
				    SX_TRACE2("Got %s",ir->answer);

			    if(need_copy) {
				if(debug>2)
				    SX_TRACE2("Copy %d bytes", need_copy);
				bcopy(p, rbuff, need_copy);
			    } else {
				n = read(ir->s, rbuff, sizeof(rbuff));
				if (n <= 0)
				    goto errlab;
			    }
			    p = strchr(rbuff, '\n');
			    *p = '\0';
			    if(debug>2) 
				    SX_TRACE2("And resulting control code is %s", rbuff);

			    if(rbuff[0]!='C') { 
				    /* smth strange.. :( */
				    return 0;
			    };
			    break;
		    case 'C': /* query successfull, no data */
		    case 'D': /* no such key */
			    break;
		    case 'E': /* multiple keys */
		    case 'F': /* another error.. */
			    return 0;
		    default:
			    return 0;
	    };

	    break;
errlab:
	    if (retries++ > 3)
		    break; 
	    SX_TRACE2("trying to reconnect");
	    if(!irrd_connect(ir))
		return 0;
   	}
	return 1;
};

int
irrd_disc(struct irrdc* ir)
{ 
	if(!ir) { 
		return 0;
	};
	if(debug>2) { 
		SX_TRACE2("sending !q");
	};
	write(ir->s, "!q\n", 3);
	close(ir->s);
	return 1;
};

int
irrd_free(struct irrdc* ir)
{
	if(!ir) return 0;

	if(ir->answer) free(ir->answer);
	free(ir);
	return 1;
};
